
from lisa import FromCoverage
import numpy as np
import argparse

def main(args):
    
    coverage_array = FromCoverage.convert_bigwig(args.bigwig_file, args.species)
    
    np.save(args.name, coverage_array)


if __name__ == "__main__":

    bw_parser = argparse.ArgumentParser()
    bw_parser.add_argument('species', choices = ['hg38','mm10'], help = 'Pileup over hg38 or mm10 genome')
    bw_parser.add_argument('bigwig_file', type = str, help = 'bigwig file to convert into coverage array')
    bw_parser.add_argument('-n', '--name', type = str, required = True, help = 'Prefix under which to save coverage array file')

    args = bw_parser.parse_args()

    main(args)


from sys import stderr
from contextlib import contextmanager
from scipy import sparse
import numpy as np
from collections import defaultdict, OrderedDict

def ragged_array_to_sparse_matrix(indices, values, col_length):
    return sparse.hstack([
        sparse.csc_matrix(
            (val_column,
            (index_column, np.zeros(len(val_column)))),
            shape = (col_length, 1)
        )
        for index_column, val_column in zip(indices, values)
    ])

def indices_list_to_sparse_array(indices_list, num_bins):
    return sparse.vstack([
        sparse.csr_matrix(
            (np.ones_like(ind), ind, [0, len(ind)]),
            shape = (1, num_bins)
        )
        for ind in indices_list
    ])

class LoadingBar:
    
    def __init__(self, label, increments, length = 25, cold_start = False):
        self.increments = increments
        self.length = length
        self.label = label
        self.progress = 0
        self.cold_start = cold_start
        
    def __str__(self):
        if self.cold_start:
            self.cold_start = False
        else:
            self.increment()
        completed_steps = int(self.progress / self.increments * self.length)
        if completed_steps >= self.length:
            return '{}: [{}]'.format(self.label, "="*completed_steps) + '\n' if self.is_finished() else ''
        else:
            return '{}: [{}>{}]'.format(self.label, "="*completed_steps, " "*(self.length - completed_steps - 1))
    
    def increment(self):
        if not self.is_finished():
            self.progress += 1
        
    def is_finished(self):
        return self.progress >= self.increments


class Log:

    def __init__(self, target = stderr, verbose = True):
        self.target = target
        self.indents = 0
        assert( isinstance(verbose, (int, bool))), 'Verbosity must be set to integer or boolean value.'
        self.verbose = verbose    

    @contextmanager
    def section(self, header):
        try:
            self.start_section(header)
            yield self
        finally:
            self.end_section()

    def start_section(self, section_header):
        self.append(section_header)
        self.indents += 1

    def end_section(self):
        self.indents -= 1 

    def append(self, text, end = '\n', update_line = False):
        linestart = '\r' if update_line else ''
        if (isinstance(self.verbose, bool) and self.verbose == True) or (isinstance(self.verbose, int) and self.indents < self.verbose):
            print(linestart + '\t'*self.indents + str(text), 
                end = '' if update_line else end, 
                file = self.target)

class LISA_Results:
    '''
Column-indexed tabular format for processing and storing rows of data. This is the
return type of the LISA test, and allows for convenient reformatting to tsv and 
pandas DataFrame

Example ::

    >>> type(results)
    lisa.core.utils.LISA_Results
    >>> results_df = pd.DataFrame(results.to_dict())
    >>> print(results.to_tsv())

    '''
    @classmethod
    def fromdict(cls, **kwargs):
        d = kwargs
        return LISA_Results(list(d.keys()), list(d.values()))

    def __init__(self, keys, columns):
        self.results_headers = keys
        self.results_rows = self._transpose_table(columns)

    def get_colnum(self, colname):
        try:
            return np.argwhere(np.array(self.results_headers) == colname)[0][0]
        except IndexError:
            raise IndexError('Column {} not in results table'.format(str(colname)))

    def get_column(self, colnames):

        column_oriented = self._transpose_table(self.results_rows)

        if type(colnames) == str:
            colnames = [colnames]
        
        returns = []
        for colname in colnames:
            colnum = self.get_colnum(colname)
            returns.append(column_oriented[colnum])

        return returns


    def sortby(self, key, add_rank = False, reverse = False):

        if isinstance(key, str):
            key = self.get_colnum(key)

        self.results_rows = sorted(self.results_rows, key = lambda col : col[key], reverse=reverse)

        if add_rank:
            row_ranks = range(1, len(self) + 1)
            try:
                self.update_column('Rank', row_ranks)
            except IndexError:
                self.add_column('Rank', row_ranks, 0)

        return self

    def __len__(self):
        return len(self.results_rows)

    @staticmethod
    def _transpose_table(table):
        return [list(l) for l in list(zip(*table))]

    def to_dict(self):
        '''
    Reformats results object to dictionary, where each key is a column, and each value is an array of the associated values. 
    By default, the rows are sorted by the "summary_p_value" field.

    Returns
    -------

    dict
        results dictionary

    Example ::

        results, metadata = lisa.predict(genelist)
        results_df = pd.DataFrame(results.to_dict())

        '''
        return dict( zip( self.results_headers, self._transpose_table(self.results_rows)))

    def update_column(self, name, data):

        colnum = self.get_colnum(name)
        
        for row, value in zip(self.results_rows, data):
            row[colnum] = value


    def add_column(self, name, data, column_num = 0):

        assert( not name in self.results_headers ), 'Column already exists'

        if column_num == -1:
            self.results_headers.append(name)
            for row, value in zip(self.results_rows, data):
                row.append(value)

        else:
            self.results_headers.insert(column_num, name)
            for row, value in zip(self.results_rows, data):
                row.insert(column_num, value)

        return self

    def subset(self, rows):

        if isinstance(rows, int):
            rows = [rows]

        return LISA_Results(self.results_headers, self._transpose_table([
            self.results_rows[row] for row in rows
        ]))

    def to_tsv(self, top_n = None):
        '''
    Reformats results object to tsv format and returns results as string. May be printed to a file for convenient storage.

    Params
    ------

    top_n : int
        Return top (n) results

    Returns
    -------

    string
        Results in tsv format

    Example ::

        results, metadata = lisa.predict(genelist)
        with open('results.tsv', 'w') as f:
            print(results.to_tsv(), file = f)

        '''
        if not top_n is None:
            output_lines = self.subset(range(top_n))
        else:
            output_lines = self
        return '\n'.join([
            '\t'.join([str(value) for value in line])
            for line in [output_lines.results_headers, *output_lines.results_rows]
        ])

    def filter_rows(self, filter_func, colname):

        colnum = self.get_colnum(colname)
        subset_rows = [
            rownum for rownum, row in enumerate(self.results_rows)
            if filter_func(row[colnum])
        ]

        return self.subset(subset_rows)


import h5py as h5
import os
import configparser
import numpy as np
from scipy import sparse

from lisa.core.utils import indices_list_to_sparse_array
from lisa.core import gene_selection
from lisa.core import genome_tools

from lisa._version import __file__ as version_file
from lisa._version import __version__
from lisa.core.utils import Log
from urllib import request, error

PACKAGE_PATH = os.path.dirname(version_file)
CONFIG_PATH = os.path.join(os.path.dirname(__file__), 'h5_config.ini')
INSTALL_PATH = os.path.join(PACKAGE_PATH, 'data')
REQURED_DATASET_VERSION = '2.0'

h5_config = configparser.ConfigParser()
h5_config.read(CONFIG_PATH)

COMPRESSION = 'gzip'

class DatasetNotFoundError(KeyError):
    pass

class DataInterface:

    _config = h5_config
    data_path = os.path.join(PACKAGE_PATH, 'data')

    @classmethod
    def get_window_bedfile_str(cls, species, window_size):

        genome = cls.load_genome(species, window_size)

        window_strs = []
        for i, window in enumerate(genome.list_windows()):
            window_strs.append(str(window) + '\t' + str(i))

        return '\n'.join(window_strs)

    @classmethod
    def get_metadata_headers(cls, technology):
        return cls._config.get('metadata', technology + '_headers').split(',')
    
    @classmethod
    def get_dataset_url(cls, species, window_size):
        return h5_config.get('lisa_params','h5_path')\
            .format(path = cls._config.get('cistrome','data_url'), species = species, 
            version = REQURED_DATASET_VERSION, window = str(window_size))

    @classmethod
    def get_dataset_path(cls, species, window_size):
        return cls._config.get('lisa_params','h5_path').format(path = cls.data_path, species = species,
                version = REQURED_DATASET_VERSION, window = str(window_size))

    #___ DATASET DOWNLOADING ____
    @classmethod
    def fetch_from_cistrome(cls, species, window_size):

        dataset_url = cls.get_dataset_url(species, window_size)

        if not os.path.isdir(cls.data_path):
            os.mkdir(cls.data_path)

        filename, _ = request.urlretrieve(dataset_url)
        os.rename(filename, cls.get_dataset_path(species, window_size))

    @classmethod
    def load_genome(cls, species, window_size):
        return genome_tools.Genome.from_file(cls._config.get('genome','genome')\
            .format(package_path = PACKAGE_PATH, species = species), 
                window_size= window_size)

    def __init__(self, species, window_size = 1000, 
        download_if_not_exists = True, make_new = False, log = None, 
        path = None, load_genes = True):

        self.species = species
        self.window_size = int(window_size)

        if log is None:
            self.log = Log()
        else:
            self.log = log

        if path is None:
            self.path = self.get_dataset_path(self.species, self.window_size)
        else:
            self.path = path

        if make_new:
            h5.File(self.path, 'w').close()
        elif not os.path.isfile(self.path):
            if download_if_not_exists and path is None:
                self.download_data()
            else:
                h5.File(self.path, 'w').close()

        #___ LOAD GENE DATA FROM PACKAGE _____
        self.genome = self.load_genome(self.species, self.window_size)

        if load_genes:
            self.load_genes()

        
    def load_genes(self):
        self.log.append('Loading gene info ...')
        self.genes = gene_selection.GeneSet.from_refseq(self._config.get('genome','genes')\
            .format(package_path = PACKAGE_PATH, species = self.species), self.genome)

        self.gene_loc_set = genome_tools.RegionSet([gene.get_tss_region() for gene in self.genes], self.genome)

        self.rp_map_locs = np.array([r.annotation.get_location() for r in self.gene_loc_set.regions])


    def get_install_path(self):
        return self.data_path

    def get_windows(self):
        return '\n'.join(
            str(r) for r in self.genome.list_windows()
        )

    # ____ RP MAP DATA _____

    @staticmethod
    def _make_basic_rp_map(gene_loc_set, region_set, decay):

        distance_matrix = gene_loc_set.map_intersects(region_set, lambda x,y : x.get_genomic_distance(y), slop_distance= 5 * decay)

        distance_matrix.data = np.power(2, -distance_matrix.data/decay)

        return distance_matrix.tocsr()

    def _make_enhanced_rp_map(self, gene_loc_set, region_set, decay):

        #make regions x exons map and exons x genes map
        try:
            indptr, indices, exons = [0],[],[]
            for locus in gene_loc_set.regions:
                new_exons = locus.annotation.get_exon_regions()
                exons.extend(new_exons)
                indices.extend(range(indptr[-1], indptr[-1] + len(new_exons)))
                indptr.append(indptr[-1] + len(new_exons))

            exon_gene_map = sparse.csc_matrix((np.ones(len(exons)), indices, indptr), shape = (len(exons), len(gene_loc_set.regions)))
            
            exons = genome_tools.RegionSet(exons, self.genome)
            region_exon_map = region_set.map_intersects(exons, distance_function = lambda x,y : x.overlaps(y, min_overlap_proportion=0.4),slop_distance=0) #REGIONS X EXONS

            region_exon_map = region_exon_map.dot(exon_gene_map).astype(np.bool)

            not_exon_promoter = 1 - region_exon_map.sum(axis = 1).astype(np.bool)

            basic_rp_map = self._make_basic_rp_map(gene_loc_set, region_set, decay)

            enhanced_rp_map = basic_rp_map.transpose().multiply(not_exon_promoter) + region_exon_map

            return enhanced_rp_map.transpose()

        except Exception as err:
            print(repr(err))
            return region_exon_map, exon_gene_map


    def build_binned_rp_map(self, style, rp_decay):

        region_set = genome_tools.RegionSet(list(self.genome.list_windows()), self.genome)

        if style == 'basic':
            return self._make_basic_rp_map(self.gene_loc_set, region_set, rp_decay)
        elif style == 'enhanced':
            return self._make_enhanced_rp_map(self.gene_loc_set, region_set, rp_decay)
        else:
            NotImplementedError()

    @staticmethod
    def set_attributes(dataset, attr_dict):
        for key, value in attr_dict.items():
            dataset.attrs[key] = value

    def get_rp_map_shape(self):
        return (len(self.genes), len(self.genome))

    def add_rp_map(self, style, rp_map):

        assert(rp_map.shape == self.get_rp_map_shape()), \
            'RP map must be of shape (num genes, num bins): ' + str(self.get_rp_map_shape())

        rp_map_path = self._config.get('rp_map','rp_map').format(style = style)

        rp_map = rp_map.tocsr()

        with h5.File(self.path, 'a') as data:

            if rp_map_path in data:
                del data[rp_map_path]

            group = data.create_group(rp_map_path)

            group.create_dataset('indptr', data = rp_map.indptr, dtype = np.int32, compression=COMPRESSION)
            group.create_dataset('indices', data = rp_map.indices, dtype = np.int32, compression=COMPRESSION)
            group.create_dataset('data', data = rp_map.data, dtype = np.float32, compression=COMPRESSION)

            self.set_attributes(group,dict(shape = rp_map.shape))

    def get_rp_maps(self):
        
        try:
            with h5.File(self.path, 'a') as data:
                return list(data['rp_maps'].keys())
        except KeyError:
            return []

    def get_rp_map(self, style):

        rp_map_path = self._config.get('rp_map','rp_map').format(style = style)

        with h5.File(self.path, 'r') as data:

            try:
                group = data[rp_map_path]

                rp_map = sparse.csr_matrix(
                    (group['data'][...], group['indices'][...], group['indptr'][...]), shape = group.attrs['shape']
                )
            except KeyError:
                raise DatasetNotFoundError(rp_map_path)
        
        return rp_map

    #___ BIN PROJECTION _____

    def check_bin_map_unique(self, bin_map):
        return len(np.unique(bin_map)) == len(bin_map)


    def project_indices(self, indices, bin_map):

        input_hits = sparse.csc_matrix(
            (np.ones_like(indices), indices, [0, len(indices)]),
        )

        input_hits = self.project_sparse_matrix(input_hits, bin_map, None)

        return input_hits.tocoo().row

    @staticmethod
    def project_array(arr, bin_map, num_bins):
        #assert(check_bin_map_unique(bin_map[:,0]) and check_bin_map_unique(bin_map[:,1])), 'To project array, bin_map must have all one-to-one mappings'
        new_arr = np.zeros(num_bins)

        new_arr[bin_map[:,1]] = arr[bin_map[:,0]]

        return new_arr

    @staticmethod
    def project_sparse_matrix(input_hits, bin_map, num_bins, binarize = False):

        index_converted = input_hits.tocsc()[bin_map[:,0], :].tocoo()

        input_hits = sparse.coo_matrix(
            (index_converted.data, (bin_map[index_converted.row, 1], index_converted.col)), 
            shape = (num_bins, input_hits.shape[1]) if not num_bins is None else None 
        ).tocsr()

        if binarize:
            input_hits.data = np.ones_like(input_hits.data)

        return input_hits


    #___ BINDING FACTOR DATA _____
    def get_factor_hit_path(self, technology, dataset_id):
        return self._config.get('factor_binding','hits').format(technology = technology, dataset_id = dataset_id)

    def get_metadata(self, attributes, technology, dataset_id):
        return {dataset_id : {key : attributes[key] for key in self.get_metadata_headers(technology)}}

    def transpose_metadata(self, metadata, technology):

        headers = self.get_metadata_headers(technology)
        sample_ids = list(metadata.keys())

        return {'sample_id' : sample_ids, **{key : [metadata[sample][key] for sample in sample_ids] for key in headers}}

    def add_binding_data(self, technology, dataset_id, hit_bins, **metadata):

        hits_path = self.get_factor_hit_path(technology, dataset_id)

        with h5.File(self.path, 'a') as data:
            if hits_path in data:
                del data[hits_path]

            hits = data.create_dataset(hits_path, data = np.array(hit_bins), dtype = np.int32, compression=COMPRESSION)
            self.set_attributes(hits, metadata)

    def get_binding_dataset(self, technology, dataset_id):

        metadata_headers = self.get_metadata_headers(technology)

        with h5.File(self.path, 'r') as data:
            
            factor_dataset_path = self.get_factor_hit_path(technology, dataset_id)
            
            try:
                hit_bins = np.array(data[factor_dataset_path][...])

                attributes = data[factor_dataset_path].attrs
            except KeyError:
                raise DatasetNotFoundError(factor_dataset_path)

            metadata = self.get_metadata(attributes, technology, dataset_id)
        
        return hit_bins, metadata

    def get_binding_data(self, technology):
        
        with h5.File(self.path, 'r') as data:

            dataset_ids = list(data[self._config.get('factor_binding','root').format(technology=technology)].keys())

            indices = []
            metadata = dict()
            for dataset_id in dataset_ids:
                hit_bins, sample_meta = self.get_binding_dataset(technology, dataset_id)

                metadata.update(sample_meta)
                indices.append(hit_bins)

        hits_matrix = indices_list_to_sparse_array(indices, len(self.genome))

        return hits_matrix.transpose(), np.array(dataset_ids), self.transpose_metadata(metadata, technology)

    def remove_binding_dataset(self, technology, dataset_id):

        factor_dataset_path = self.get_factor_hit_path(technology, dataset_id)

        with h5.File(self.path, 'a') as data:
            del data[factor_dataset_path]

    def list_binding_datasets(self, technology):

        try:
            with h5.File(self.path, 'r') as data:

                dataset_ids = list(data[self._config.get('factor_binding','root').format(technology=technology)].keys())

            return dataset_ids

        except KeyError:
            return []
        

    #____ PROFILE DATA _____
    def add_profile_data(self, technology, dataset_id, profile, rp_maps, rp_map_styles, 
        norm_depth = 1e5, **metadata):
        
        assert(len(rp_maps) == len(rp_map_styles))

        profile_path = self._config.get('profiles','profile').format(technology = technology, dataset_id = dataset_id)

        profile = np.array(profile)
        if len(profile.shape) == 1:
            profile = profile[:,np.newaxis]
        assert(len(profile.shape) == 2)
        assert(profile.shape[0] == self.genome.num_windows_in_genome())
        
        if not norm_depth is None:
            profile = profile/profile.sum() * norm_depth

        with h5.File(self.path, 'a') as data:

            if profile_path in data:
                del data[profile_path]
            
            hits = data.create_dataset(profile_path, data = profile, dtype = np.float16, compression=COMPRESSION)
            self.set_attributes(hits, metadata)

            for rp_map, style in zip(rp_maps, rp_map_styles):

                rp_matrix_path = self._config.get('profiles','rp_matrix_col').format(technology = technology, style = style, dataset_id = dataset_id)

                if rp_matrix_path in data:
                    del data[rp_matrix_path]

                rp_matrix_col = data.create_dataset(rp_matrix_path, data = rp_map.dot(profile), dtype = np.float32, compression=COMPRESSION)
                self.set_attributes(rp_matrix_col, metadata)

    def remove_profile(self, technology, dataset_id):

        profile_path = self._config.get('profiles','profile').format(technology = technology, dataset_id = dataset_id)

        with h5.File(self.path, 'a') as data:
            del data[profile_path]
            
            for style in self.get_rp_maps():
                rp_matrix_col_path = self._config.get('profiles','rp_matrix_col').format(technology = technology, style = style, dataset_id = dataset_id)
                del data[rp_matrix_col_path]


    def get_profile(self, technology, dataset_id):

        profile_path = self._config.get('profiles','profile').format(technology = technology, dataset_id = dataset_id)

        with h5.File(self.path, 'r') as data:

            try:
                profile = np.array(data[profile_path][...])

                attributes = data[profile_path].attrs
            except KeyError:
                raise DatasetNotFoundError(profile_path)

            metadata = self.get_metadata(attributes, technology, dataset_id)

        return profile, metadata

    def list_profiles(self, technology):

        profiles_dir = self._config.get('profiles','root').format(technology = technology)

        try:
            with h5.File(self.path, 'r') as data:
                
                dataset_ids = list(data[profiles_dir].keys())

            return dataset_ids
        except KeyError:
            return []

    def get_rp_matrix(self, technology, style):

        with h5.File(self.path, 'r') as data:
            
            rp_matrix_dir = self._config.get('profiles','rp_matrix').format(technology = technology, style = style)

            dataset_ids = list(data[rp_matrix_dir].keys())

            slices = []
            for _id in dataset_ids:
                slices.append(
                    np.array(data[rp_matrix_dir][_id][...])
                )

        return np.concatenate(slices, axis = 1), np.array(dataset_ids)

    def download_data(self):

        with self.log.section('Grabbing {} data (~15 minutes):'.format(self.species)):
            
            self.log.append('Downloading from database ...')

            try:
                self.fetch_from_cistrome(self.species, self.window_size)
            except error.URLError as err:
                raise AssertionError('ERROR: Cannot connect to cistrome.org for data (usually due to security settings on some servers)!\nView github pages for manual dataset install instructions.')

            self.log.append('Done')


import numpy as np
from scipy import stats, sparse
import os
import warnings

def get_delta_RP(profile, binding_data, rp_map):
    '''
    profile: chromatin profile, bin x 1 array of accessibility at each genomic region
    rp_map: sparse bin x gene matrix mapping a genomic region to a gene based on RP proximity
    binding_data: bin x dataset matrix, with binary hits to show TF binding occurs in a location

    the line below defines the mapping of chromatin accessibility at TF binding sites to genes.

    returns: gene x TF x 1 matrix of delta RPs
    '''
    return np.array(rp_map.dot(binding_data.astype(np.bool).multiply(profile.reshape((-1,1)))).todense())

#distributes arguments for get_delta_RP function, to be used by multiprocessing module
def delta_RP_wrapper(x):
    return get_delta_RP(*x)[:,np.newaxis,:]

def mannu_test_function(x):
    query, background = x
    try:
        return stats.mannwhitneyu(query, background, alternative = 'greater')
    #if all values in query and background are equal (no knockouts), throws value error
    except ValueError:
        #catch, return none for test statistic, 1.0 for p-val
        return (None, 1.0)

def get_deltaRP_activation(rp_0, rp_knockout):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        deltaX = np.sqrt(np.where(rp_0 == 0, 0, rp_knockout/rp_0))
    
    return deltaX

def transform_RP(x):
    return np.sqrt(x)

'''def transform_RP(x):
    return np.log(x + 1)

def get_deltaRP_activation(rp_0, rp_knockout):
    return -1 * np.log2(1 - rp_knockout/(rp_0 + 0.01))'''

class LISA_RP_Assay:

    def __init__(self, *, technology, data_interface, log, rp_map, factor_binding, factor_dataset_ids):
        self.data_interface = data_interface
        self.technology = technology
        self.log = log
        self.rp_map = rp_map
        self.factor_binding = factor_binding
        self.factor_dataset_ids = factor_dataset_ids
    
    @staticmethod
    def make_subset_rp_map(*,rp_map, factor_binding, gene_mask, accessibility):
        
        bin_mask = np.squeeze(np.array(rp_map[gene_mask, : ].tocsc().sum(axis = 0) > 0))
        #subset rp_map and factor hits on bins with RP > 0
        subset_factor_binding = factor_binding[bin_mask, :]
        
        subset_rp_map = rp_map[gene_mask, :][:, bin_mask]

        subset_accessibility = accessibility[bin_mask, :]
        
        return subset_accessibility, subset_factor_binding, subset_rp_map

    @staticmethod
    def make_rp_matrix(*, profiles, rp_map):

        rp_matrix = rp_map.dot(profiles)

        return rp_matrix

    def get_delta_RP_p_value(self, gene_TF_scores, label_vector):
        '''
        gene_TF_scores: gene x TF, model output of delta-RP matrix. more purturbation of genes of interest correspond with higher delta regulation score
        '''
        #seperate matrix into query and background sets
        query_delta = gene_TF_scores[label_vector.astype(np.bool)]
        background_delta = gene_TF_scores[~label_vector.astype(np.bool)]

        #for each TF, calculate p-value of difference in delta-R distributions between query and background genes
        test_parameters = list(zip(query_delta.T, background_delta.T))

        p_vals = [
            mannu_test_function((q,b)) for q,b in test_parameters
        ]

        _, p_values = list(zip(*p_vals))

        return p_values

    @staticmethod
    def get_delta_reg_score_matrix(p_values, delta_reg_scores, label_vector, top_n = 100):

        label_vector = label_vector.astype(np.bool)
        
        most_significant_datasets = np.argsort(p_values)[:top_n]
        narrowed_reg_scores = delta_reg_scores[:, most_significant_datasets]

        return narrowed_reg_scores[label_vector], narrowed_reg_scores[~label_vector].mean(axis = 0, keepdims = True),  most_significant_datasets


    def predict(self, gene_mask, label_vector, *args, data_object = None, debug = False, **kwargs):
        raise NotImplementedError()


import unittest
from lisa.core import genome_tools
import numpy as np

class TestRegionObject(unittest.TestCase):

    def test_invalid_region(self):
        with self.assertRaises(AssertionError):
            genome_tools.Region('chr1',10, 5)

    def test_overlap_different_chr(self):
        self.assertFalse(
            genome_tools.Region('chr1',5,10).overlaps(genome_tools.Region('chr2',5,10))
        )

    def test_overlap_negative(self):
        self.assertFalse(
            genome_tools.Region('chr1',5,10).overlaps(genome_tools.Region('chr1',20,25))
        )

    def test_overlap_negative_abutted(self):
        self.assertFalse(
            genome_tools.Region('chr1',5,10).overlaps(genome_tools.Region('chr1',10,20))
        )

    def test_overlap_any(self):
        self.assertTrue(
            genome_tools.Region('chr1',5,10).overlaps(genome_tools.Region('chr1',8,15))
        )

        self.assertFalse(
            genome_tools.Region('chr1',1,11).overlaps(genome_tools.Region('chr1',9,15), min_overlap_proportion=0.4)
        )

        self.assertTrue(
            genome_tools.Region('chr1',1,11).overlaps(genome_tools.Region('chr1',6,15), min_overlap_proportion=0.4)
        )

    def test_distance_function(self):
        self.assertEqual(
            genome_tools.Region('chr1',20, 30).get_genomic_distance(genome_tools.Region('chr1', 120, 130)), 100)

    def test_regions_equal(self):
        self.assertTrue(
            genome_tools.Region('chr1',10,15) == genome_tools.Region('chr1', 10, 15)
        )

class TestGenomeObject(unittest.TestCase):

    def setUp(self):
        self.genome = genome_tools.Genome(['chr1','chr2','chr3'], [450, 375, 600], window_size=100)
        
        self.unsorted_genome = genome_tools.Genome(['chr1','chr3','chr2'],[200, 300, 250], window_size= 100, _sort=False)
        self.sorted_genome = genome_tools.Genome(['chr1','chr3','chr2'],[200, 400, 250], window_size= 100, _sort=True)
        
        self.correct_mapping = np.array([
            (0,0),
            (1,1),
            (2,5), 
            (3,6),
            (4,7),
            (5,2),
            (6,3),
            (7,4),
        ])

    def test_genome_genome_mapping(self):

        self.assertTrue(
            np.all(self.unsorted_genome.map_genomes(self.sorted_genome) == self.correct_mapping)
        )

    def test_indptr(self):
        self.assertEqual(
            tuple(self.genome.indptr), (0,5,9,15)
        )
    
    def test_region_check(self):
        with self.assertRaises(genome_tools.BadRegionError):
            self.genome.check_region(genome_tools.Region('chr4',10,20))

        with self.assertRaises(genome_tools.BadRegionError):
            self.genome.check_region(genome_tools.Region('chr1',-1, 20))

        with self.assertRaises(genome_tools.BadRegionError):
            self.genome.check_region(genome_tools.Region('chr2', 300, 400))

    def test_chromlen(self):
        self.assertEqual(
            self.genome.get_chromlen('chr2'), 375
        )

    def test_get_num_windows(self):
        self.assertEqual(
            self.genome.get_num_windows(self.genome.get_chromlen('chr1'), self.genome.window_size), 5
        )

        self.assertEqual(
            self.genome.num_windows_in_genome(), 15
        )

    def test_get_window_from_position(self):
        region, window_idx = self.genome.get_window_from_position('chr2', 250)
        self.assertTrue(
            region == genome_tools.Region('chr2', 200, 300)
        )

        next_region, next_window_idx = self.genome.get_next_window(region)
        self.assertTrue(
            next_region == genome_tools.Region('chr2',300,375)
        )

        self.assertTrue(
            window_idx == (next_window_idx - 1)
        )

    def test_get_window_idx(self):
        self.assertTrue(
            self.genome.get_region(0)[0] == genome_tools.Region('chr1', 0, 100)
        )

        self.assertTrue(
            self.genome.get_region(4)[0] == genome_tools.Region('chr1',400,450)
        )

        self.assertTrue(
            self.genome.get_region(5)[0] == genome_tools.Region('chr2',0,100)
        )

        self.assertTrue(
            self.genome.get_region(14)[0] == genome_tools.Region('chr3',500,600)
        )

        with self.assertRaises(AssertionError):
            self.genome.get_region(15)


class TestRegionSet(unittest.TestCase):

    def setUp(self):
        self.genome = genome_tools.Genome(['chr1','chr2','chr3'], [450, 375, 600], window_size=50)
        self.scrambled_genome = genome_tools.Genome(['chr1','chr3','chr2'], [450, 600, 375], window_size=50)

        self.regions1A = [
            genome_tools.Region('chr1',20,40),
            genome_tools.Region('chr1',30,60),
            genome_tools.Region('chr1',210,230),
            genome_tools.Region('chr2',100,150),
            genome_tools.Region('chr2',220,233),
            genome_tools.Region('chr3',430,450),
        ]
        self.regions1B = [
            genome_tools.Region('chr1',20,40),
            genome_tools.Region('chr1',30,60),
            genome_tools.Region('chr1',210,230),
            genome_tools.Region('chr2',100,150),
            genome_tools.Region('chr2',220,233),
            genome_tools.Region('chr3',430,450),
        ]

        self.auto_distancing_truth = np.array(
            [[0,15,0,0,0,0],
             [15,0,175,0,0,0],
             [0,175,0,0,0,0],
             [0,0,0,0,101,0],
             [0,0,0,101,0,0],
             [0,0,0,0,0,0]
            ]
        )

        self.m2m_map_truth = np.array([
            (0,0),
            (0,1),
            (1,1),
            (4,2),
            (11,3),
            (13,4),
            (25,5)
        ])

    def test_auto_distancing(self):
        distance_matrix = genome_tools.RegionSet(self.regions1A, self.genome).map_intersects(genome_tools.RegionSet(self.regions1B, self.genome), 
            lambda x,y : x.get_genomic_distance(y), slop_distance=75)

        self.assertTrue(
            np.all(np.array(distance_matrix.todense()).astype(int) == self.auto_distancing_truth)
        )

    def test_auto_distancing_scrambled(self):

        distance_matrix = genome_tools.RegionSet(self.regions1A, self.scrambled_genome)\
            .map_intersects(genome_tools.RegionSet(self.regions1B, self.scrambled_genome), 
            lambda x,y : x.get_genomic_distance(y), slop_distance=75)

        self.assertTrue(
            np.all(np.array(distance_matrix.todense()).astype(int) == self.auto_distancing_truth)
        )

    def test_genome_bin_mapping(self):

        m2m_map = genome_tools.RegionSet(self.regions1A, self.genome)\
            .map_genomic_windows(min_window_overlap_proportion=0.0, regions_to_bins=False)

        self.assertTrue(
            np.all(m2m_map == self.m2m_map_truth)
        )

if __name__ == '__main__':
    unittest.main()



import lisa
from lisa import FromRegions, FromGenes, FromCoverage
import argparse
from lisa.cli.cli import oneshot_parser, multi_parser, regions_parser, coverage_parser
import argparse
import os
from argparse import HelpFormatter
from lisa.cli.cli import RstFormatter
from lisa.core.utils import Log

def new_format_help(self):

    formatter = RstFormatter(prog = self.prog)

    # description
    formatter.add_text(self.description)

    # usage
    formatter.add_usage(self.usage, self._actions,
                        self._mutually_exclusive_groups)

    # positionals, optionals and user-defined groups
    for action_group in self._action_groups:
        formatter.start_section(action_group.title)
        formatter.add_text(action_group.description)
        formatter.add_arguments(action_group._group_actions)
        formatter.end_section()

    # epilog
    formatter.add_text(self.epilog)

    # determine help from format above
    return formatter.format_help()

def make_cli_page(parsers):

    docs = lisa.cli.cli.__doc__

    docs += '\n\n'.join([new_format_help(m) for m in parsers]).replace('make_docs.py', 'lisa')

    return docs


def make_api_page(modules):

    docs = lisa.__doc__

    docs += '\n\n'.join([m.get_docs() for m in modules])

    return docs

if __name__ == "__main__":
    
    api_docs = make_api_page([FromGenes, FromRegions, FromCoverage])

    with open('python_api.rst', 'w') as f:
        print(api_docs, file = f)

    cli_docs = make_cli_page([oneshot_parser, multi_parser, regions_parser, coverage_parser])

    with open('cli.rst', 'w') as f:
        print(cli_docs, file = f)


'''
********
Lisa CLI
********

Installing LISA using pip or conda adds the "lisa" command to your path. LISA's functionality is divided into three main subcommands:

* `lisa oneshot`_ : one genelist
* `lisa multi`_ : multiple genelists
* `lisa regions`_ : one genelist and a list of regions

Which are used depending on the evidence you have on hand. 

See the `User Guide <user_guide.rst>`_ for more usage information.
See the `Python API <python_api.rst>`_ for more in-depth description of tests and parameters.

'''

from lisa import FromRegions, FromGenes, FromCoverage
from lisa.core.utils import Log
from lisa.core.lisa_core import DownloadRequiredError
from lisa.core.data_interface import DatasetNotFoundError, INSTALL_PATH
from lisa._version import __version__
import configparser
import argparse
import os
import sys
import json
from collections import defaultdict
from shutil import copyfile
import lisa.cli.test_cli as tests
from shutil import copyfile
import numpy as np

from lisa.lisa_public_data.genes_test import _config as public_config
from lisa.lisa_user_data.regions_test import _config as user_config
from lisa.core.io import parse_deseq_file

#____COMMAND LINE INTERFACE________

INSTANTIATION_KWARGS = ['isd_method','verbose','assays', 'rp_map']
PREDICTION_KWARGS = ['background_list','num_background_genes','background_strategy', 'seed']

def extract_kwargs(args, keywords):
    return {key : vars(args)[key] for key in keywords}

def is_valid_prefix(prefix):

    if '/' in prefix:
        if os.path.isdir(prefix) or os.path.isfile(prefix) or os.path.isdir(os.path.dirname(prefix)):
            return prefix
        else:
            raise argparse.ArgumentTypeError('{}: Invalid file prefix.'.format(prefix))
    else:
        return prefix

def save_results(args, results, metadata):

    if args.save_metadata:
        if args.output_prefix:
            metadata_filename = args.output_prefix + '.metadata.json' 
        else:
            metadata_filename = os.path.basename(args.query_list.name) + '.metadata.json'

        with open(metadata_filename, 'w') as f:
            f.write(json.dumps(metadata, indent=4))

    if not args.output_prefix is None:
        with open(args.output_prefix + '.lisa.tsv', 'w') as f:
            f.write(results.to_tsv())
    else:
        print(results.to_tsv())

def lisa_oneshot(args):

    try:
        args.background_list = args.background_list.readlines()
    except AttributeError:
        pass

    results, metadata = FromGenes(args.species, **extract_kwargs(args, INSTANTIATION_KWARGS)).predict(args.query_list.readlines(), **extract_kwargs(args, PREDICTION_KWARGS))
    
    save_results(args, results, metadata)


def lisa_regions(args):

    try:
        args.background_list = args.background_list.readlines()
    except AttributeError:
        pass

    if not args.macs_xls:
        fn = FromRegions.using_bedfile
    else:
        fn = FromRegions.using_macs_output

    results, metadata = fn(args.species, args.query_genes, args.regions, rp_map = args.rp_map,
        rp_decay=args.rp_decay, isd_method=args.isd_method, background_list=args.background_list,
        background_strategy=args.background_strategy, num_background_genes = args.num_background_genes,
        seed=args.seed, header = args.header)
    
    save_results(args, results, metadata)

def lisa_coverage(args):

    try:
        args.background_list = args.background_list.readlines()
    except AttributeError:
        pass

    results, metadata = FromCoverage.using_bigwig(args.species, args.query_genes, args.bigwig_path, rp_map = args.rp_map,
        isd_method=args.isd_method, background_list=args.background_list,
        background_strategy=args.background_strategy, num_background_genes = args.num_background_genes,
        seed=args.seed)
    
    save_results(args, results, metadata)


def save_and_get_top_TFs(args, query_name, results, metadata):

    with open(args.output_prefix + query_name + '.lisa.tsv', 'w') as f:
        f.write(results.to_tsv())

    if args.save_metadata:
        with open(args.output_prefix + query_name + '.metadata.json', 'w') as f:
            f.write(json.dumps(metadata, indent=4))

    top_TFs = results.to_dict()['factor']

    return list(set(top_TFs[:10]))

def print_results_multi(results_summary):
    print('Sample\tTop Regulatory Factors:')
    for result_line in results_summary:
        print(result_line[0], ', '.join(result_line[1]), sep = '\t')

class MultiError(Exception):
    pass

def lisa_multi(args):

    log = Log(target = sys.stderr, verbose = args.verbose)
    lisa = FromGenes(args.species, **extract_kwargs(args, INSTANTIATION_KWARGS), log = log)

    query_dict = {os.path.basename(query.name) : query.readlines() for query in args.query_lists}

    results_summary = []
    all_passed = True
    for query_name, query_list in query_dict.items():
    
        with log.section('Modeling {}:'.format(str(query_name))):
            try: 
                results, metadata = lisa.predict(query_list, **extract_kwargs(args, PREDICTION_KWARGS))

                top_TFs_unique = save_and_get_top_TFs(args, query_name, results, metadata)
            
                results_summary.append((query_name, top_TFs_unique))
            
            except AssertionError as err:
                all_passed = False
                log.append('ERROR: ' + str(err))

    print_results_multi(results_summary)

    if not all_passed:
        raise MultiError('One or more genelists raised an error')

def lisa_deseq(args):

    log = Log(target = sys.stderr, verbose = args.verbose)
    lisa = FromGenes(args.species, **extract_kwargs(args, INSTANTIATION_KWARGS), log = log)

    up_genes, down_genes = parse_deseq_file(args.deseq_file, lfc_cutoff = args.lfc_cutoff, 
        pval_cutoff= args.pval_cutoff, sep = args.sep)

    results_summary = []
    all_passed = True
    for prefix, query_list in zip(['up-regulated', 'down-regulated'], [up_genes, down_genes]):
    
        with log.section('Modeling {}:'.format(str(prefix))):
            try: 
                results, metadata = lisa.predict(query_list, **extract_kwargs(args, PREDICTION_KWARGS))

                top_TFs_unique = save_and_get_top_TFs(args, prefix, results, metadata)
            
                results_summary.append((prefix, top_TFs_unique))
            
            except AssertionError as err:
                all_passed = False
                log.append('ERROR: ' + str(err))

    print_results_multi(results_summary)

    if not all_passed:
        raise MultiError('One or more genelists raised an error')

def confirm_file(arg):
    if os.path.isfile(arg):
        return arg
    else:
        raise argparse.ArgumentTypeError('ERROR: {} is not a valid file'.format(str(arg)))

def run_tests(args):

    if not args.skip_oneshot:
        tests.test_oneshot(args.test_genelist, args.background_genelist)

    tests.test_multi(args.genelists)
        
def build_common_args(parser):
    parser.add_argument('--seed', type = int, default = 2556, help = 'Random seed for gene selection. Allows for reproducing exact results.')
    parser.add_argument('--use_motifs', action = 'store_const', const = 'motifs', default='chipseq',
        dest = 'isd_method', help = 'Use motif hits instead of ChIP-seq peaks to represent TF binding (only recommended if TF-of-interest is not represented in ChIP-seq database).')
    parser.add_argument('--save_metadata', action = 'store_true', default = False, help = 'Save json-formatted metadata from processing each gene list.')

def build_from_genes_args(parser, add_assays = True):
    #parser.add_argument('-c','--cores', required = True, type = int)
    if add_assays:
        parser.add_argument('-a','--assays',nargs='+',default=['Direct','H3K27ac','DNase'], choices=['Direct','H3K27ac','DNase'], help = 'Which set of insilico-deletion assays to run.')
    parser.add_argument('--rp_map_style', dest = 'rp_map', choices=public_config.get('lisa_params','rp_map_styles').split(','),
        default= public_config.get('lisa_params','rp_map_styles').split(',')[0], help = 'Which style of rp_map to assess influence of regions on genes. "basic" is stricly distance-based, while "enhanced" masks the exon and promoter regions of nearby genes.')

def build_multiple_lists_args(parser):
    parser.add_argument('-o','--output_prefix', required = True, type = is_valid_prefix, help = 'Output file prefix.')
    parser.add_argument('-v','--verbose',type = int, default = 2)
    parser.add_argument('-b','--num_background_genes', type = int, default = public_config.get('lisa_params', 'background_genes'),
        help = 'Number of sampled background genes to compare to user-supplied genes. These genes are selection from other gene lists.')
    parser.add_argument('--random_background', action = 'store_const', const = 'random', default = 'regulatory', dest = 'background_strategy', help = 'Use random background selection rather than "regulatory" selection.')
    
def build_one_list_args(parser, default_background_strategy = 'regulatory'):
    parser.add_argument('-o','--output_prefix', required = False, type = is_valid_prefix, help = 'Output file prefix. If left empty, will write results to stdout.')
    parser.add_argument('--background_strategy', choices = public_config.get('lisa_params', 'background_strategies').split(','),
        default = default_background_strategy,
        help = """Background genes selection strategy. LISA samples background genes to compare to user\'s genes-of-interest from a diverse
        regulatory background (regulatory - recommended), randomly from all genes (random), or uses a user-provided list (provided).
        """)
    background_genes_group = parser.add_mutually_exclusive_group()
    background_genes_group.add_argument('--background_list', type = argparse.FileType('r', encoding = 'utf-8'), required = False,
        help = 'user-supplied list of backgroung genes. Used when --background_strategy flag is set to "provided"')
    background_genes_group.add_argument('-b','--num_background_genes', type = int, default = public_config.get('lisa_params', 'background_genes'),
        help = 'Number of sampled background genes to compare to user-supplied genes')
    parser.add_argument('-v','--verbose',type = int, default = 4)

def build_deseq_args(parser):
    parser.add_argument('deseq_file', type = confirm_file, help = 'DEseq differential expression output file. Will be parsed for differentially up and down-regulated genes.')
    parser.add_argument('-lfc','--lfc_cutoff', type = float, default = 2, help = 'Log2 fold-change cutoff. For up-regulated genes, must have LFC > cutoff. For down-regulated genes, less than -1 * cutoff. Default of 1 means genes must be up or down-regulated by a factor of 2 to be included in query.')
    parser.add_argument('-p','--pval_cutoff', type = float, default = 0.1, help = 'Adjusted p-value cutoff. Gene must have pval below cutoff to be a query gene.')
    parser.add_argument('--sep', type = str, default='\t', help = 'Field separator for DESeq output file.')
   

class RstFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawTextHelpFormatter):
    pass

parser = argparse.ArgumentParser(
        formatter_class=RstFormatter,
        description =
"""
Lisa: inferring transcriptional regulators through integrative modeling of public chromatin accessibility and ChIP-seq data
https://genomebiology.biomedcentral.com/articles/10.1186/s13059-020-1934-6
X. Shirley Liu Lab, 2020
""")
parser.add_argument('--version', action = 'version', version = __version__)
subparsers = parser.add_subparsers(help = 'commands')

#__ LISA oneshot command __
oneshot_parser = subparsers.add_parser('oneshot', formatter_class=RstFormatter, description = '''
lisa oneshot
------------

You have:

* one genelist

Use LISA to infer influential TFs from one gene list, with background epigenetic landscape modeled using public data. 
If you have multiple lists, this option will be slower than using "multi" due to data-loading time. \n

Example::

    $ lisa oneshot hg38 ./genelist.txt -b 501 --seed=2556 --save_metadata > results.tsv

''')
oneshot_parser.add_argument('species', choices = ['hg38','mm10'], help = 'Find TFs associated with human (hg38) or mouse (mm10) genes')
oneshot_parser.add_argument('query_list', type = argparse.FileType('r', encoding = 'utf-8'), help = 'user-supplied gene lists. One gene per line in either symbol or refseqID format')
build_one_list_args(oneshot_parser)
build_from_genes_args(oneshot_parser)
build_common_args(oneshot_parser)
oneshot_parser.set_defaults(func = lisa_oneshot)

deseq_parser = subparsers.add_parser('deseq', formatter_class = RstFormatter, description = '''
lisa deseq
----------

You have:

* RNA-seq differential expression results from DESeq2

Use LISA to infer influential TFs given differentially expressed genes found using DESeq2. Will seperate up-regulated and down-regulated genes into their own LISA tests.

Example::

    $ lisa deseq hg38 ./deseq_results.tsv -o deseq/ -b 501 --seed=2556 --save_metadata

''')
deseq_parser.add_argument('species', choices = ['hg38','mm10'], help = 'Find TFs associated with human (hg38) or mouse (mm10) genes')
build_deseq_args(deseq_parser)
build_multiple_lists_args(deseq_parser)
build_from_genes_args(deseq_parser)
build_common_args(deseq_parser)
deseq_parser.set_defaults(func = lisa_deseq, background_list = None)

#__ LISA multi command __
multi_parser = subparsers.add_parser('multi', formatter_class=RstFormatter, description = '''
lisa multi
----------

You have:

* multiple genelists

Use LISA to infer influential TFs from multiple lists. This function processes each genelist independently in the same manner as the "oneshot" command, but reduces data loading time. Useful when performing 
the test on up and down-regulated genes from multiple RNA-seq clusters.

Example::

    $ lisa multi hg38 ./genelists/*.txt -b 501 -o ./results/

''')
multi_parser.add_argument('species', choices = ['hg38','mm10'], help = 'Find TFs associated with human (hg38) or mouse (mm10) genes')
multi_parser.add_argument('query_lists', type = argparse.FileType('r', encoding = 'utf-8'), nargs = "+", help = 'user-supplied gene lists. One gene per line in either symbol or refseqID format')
build_multiple_lists_args(multi_parser)
build_from_genes_args(multi_parser)
build_common_args(multi_parser)
multi_parser.set_defaults(func = lisa_multi, background_list = None)

from argparse import SUPPRESS
#____ LISA regions command ____
regions_parser = subparsers.add_parser('regions', formatter_class=RstFormatter, add_help = False, description = '''
lisa regions
------------

You have:

* one genelist
* regions (250 - 1000 bp wide) of interest related to that list
* optional: a positive score/weight associated with each region (you may pass zero-weight regions, but they do not affect the test and will be filtered out)

Use LISA to infer TF influence on your geneset, but provide your regions-of-interest rather than building a background epigenetic model using public data. When providing 
your own regions, LISA uses higher resolution, more precise binding data to increase the power of the test. Your regions should be between ~250 and 1000 bp in width, and the 
associated score should be positive. Scores are often read-depth at those regions, but can be any metic you think may influence gene regulation.

Example::

    $ lisa regions -r ./regions.bed -q ./genelist.txt --save_metadata > results.tsv
    $ lisa regions -r ./macs_peaks.xls -q ./genelist.txt --macs_xls > results.tsv

''')
regions_parser.add_argument('species', choices = ['hg38','mm10'], help = 'Find TFs associated with human (hg38) or mouse (mm10) genes')
regions_required = regions_parser.add_argument_group('required arguments')
regions_required.add_argument('-q', '--query_genes', required = True, type = argparse.FileType('r', encoding = 'utf-8'), help = 'user-supplied gene list. One gene per line in either symbol or refseqID format')
regions_required.add_argument('-r', '--regions', type = confirm_file, required = True, help = 'Tad-delineated bed file with columns: chr, start, end[, score]. The score column is optional. If not provided, LISA will assign each region a uniform weight.')
regions_optional = regions_parser.add_argument_group('optional arguments')
regions_optional.add_argument('--header', action = 'store_true', default=False, help = 'Bed file has header row as first row. The header row may contain ')
regions_optional.add_argument('--macs_xls', action = 'store_true', default=False, help='If provided, regions file is a MACS2 .xls output file, and the "pileup" field is taken to be the region score.')
regions_optional.add_argument('--rp_map_style', dest = 'rp_map', choices=user_config.get('lisa_params','rp_map_styles').split(','),
    default=user_config.get('lisa_params','rp_map_styles').split(',')[0])
regions_optional.add_argument('--rp_decay', type = int, default = user_config.get('lisa_params','rp_decay'),
    help = 'Distance in base-pairs in which the influence of a region on a gene decays by half. Increase for more weight on distal elements, decrease for more weight on promoter elements.')
build_one_list_args(regions_optional, default_background_strategy='all')
build_common_args(regions_optional)
regions_optional.add_argument('-h', '--help', action = 'help', default=SUPPRESS)
regions_parser.set_defaults(func = lisa_regions)

#___ LISA coverage commands _____

coverage_parser = subparsers.add_parser('coverage', formatter_class = RstFormatter, add_help = False, description = '''
lisa coverage
------------

You have:

* one genelist
* bigwig of coverage over the genome

Use LISA to infer TF influence on your geneset using your own coverage data. This test is better suited than the "regions" test when your measure produces wide peaks/areas of influence.
An example of this is H3K27ac data, which correlates with gene expression similarly to accessibility, but produces wide peaks that may span many distinct TF binding locations.

Example::

    $ lisa coverage -bw ./sample.bigwig -q ./genelist.txt --save_metadata > results.tsv

''')
coverage_parser.add_argument('species', choices = ['hg38','mm10'], help = 'Find TFs associated with human (hg38) or mouse (mm10) genes')
coverage_parser.add_argument('-q', '--query_genes', required = True, type = argparse.FileType('r', encoding = 'utf-8'), help = 'user-supplied gene list. One gene per line in either symbol or refseqID format')
coverage_parser.add_argument('-bw', '--bigwig_path', type = confirm_file, required = True, help = 'Bigwig file describing coverage over the genome.')
coverage_optional = coverage_parser.add_argument_group('optional arguments')
build_from_genes_args(coverage_optional, False)
build_one_list_args(coverage_optional, default_background_strategy='all')
build_common_args(coverage_optional)
coverage_optional.add_argument('-h', '--help', action = 'help', default=SUPPRESS)
coverage_parser.set_defaults(func = lisa_coverage)

#__ download command ___

def lisa_download(args):
    if args.command in ['oneshot','multi','coverage']:
        _class = FromGenes
    elif args.command == 'regions':
        _class = FromRegions
    else:
        raise AssertionError('Command {} not recognized'.format(args.command))
    if args.url:
        print(_class.get_dataset_url(args.species))
    else:
        _class.download_dataset(args.species)

download_data_parser = subparsers.add_parser('download', description = 'Download data from CistromeDB. Use if data recieved is incomplete or malformed.')
download_data_parser.add_argument('species', choices = ['hg38','mm10'], help = 'Download data associated with human (hg38) or mouse (mm10) genes')   
download_data_parser.add_argument('command', choices=['oneshot', 'multi', 'regions', 'coverage'], help = 'For which command to download data')
download_data_parser.add_argument('--url', action = 'store_true', help = 'Get url for data download. Does not install data.')
download_data_parser.set_defaults(func = lisa_download)

#__ install command ___

def install_data(args):
    if args.command in ['oneshot','multi','coverage']:
        _class = FromGenes
    elif args.command == 'regions':
        _class = FromRegions
    else:
        raise AssertionError('Command {} not recognized'.format(args.command))
    
    dataset_file_required = os.path.basename(_class.get_dataset_path(args.species))

    if not args.force:
        assert(dataset_file_required == os.path.basename(args.dataset)), 'The {} test requires dataset {}. Use --force to overide and install your own dataset.'.format(args.command, dataset_file_required)

    if not os.path.isdir(INSTALL_PATH):
        os.mkdir(INSTALL_PATH)
    
    if args.remove:
        os.rename(args.dataset, _class.get_dataset_path(args.species))
    else:
        copyfile(args.dataset, _class.get_dataset_path(args.species))


install_data_parser = subparsers.add_parser('install', description = 'Helper command for manually installing Lisa\'s data')
install_data_parser.add_argument('species', choices = ['hg38','mm10'], help = 'Install data associated with human (hg38) or mouse (mm10) genes')   
install_data_parser.add_argument('command', choices=['oneshot', 'multi', 'regions', 'coverage'], help = 'For which command to install data')
install_data_parser.add_argument('dataset', type = confirm_file, help = 'Path to downloaded h5 dataset')
install_data_parser.add_argument('--remove', action = 'store_true', help = 'Delete dataset after installation is complete.')
install_data_parser.add_argument('--force', action = 'store_true', help = 'Skip namecheck and install lisa custom dataset')
install_data_parser.set_defaults(func = install_data)

#____ LISA run tests command ___
test_parser = subparsers.add_parser('run-tests')
test_parser.add_argument('species', type = str, choices=['hg38','mm10'])
test_parser.add_argument('test_genelist', type = confirm_file, help = 'test genelist for oneshot command')
test_parser.add_argument('background_genelist', type = confirm_file, help = 'background genelist for oneshot command')
test_parser.add_argument('genelists', nargs = '+', type = str, help = 'genelists for testing multi and one-vs-rest commands')
test_parser.add_argument('--skip_oneshot', action='store_true')
args = parser.parse_args()
test_parser.set_defaults(func = run_tests)
    
def main():
    #____ Execute commands ___
    args = parser.parse_args()

    try:
        args.func #first try accessing the .func attribute, which is empty if user tries ">>>lisa". In this case, don't throw error, display help!
    except AttributeError:
        print(parser.print_help(), file = sys.stderr)
    else:
        try:
            args.func(args)
        except (AssertionError, DownloadRequiredError, DatasetNotFoundError, MultiError) as err:
            print('ERROR: ' + str(err), file = sys.stderr)
            sys.exit(1)



import numpy as np
from collections import Counter
from scipy import sparse

class BadRegionError(Exception):
    pass

class NotInGenomeError(BadRegionError):
    pass

class Region:

    @classmethod
    def read_bedfile(cls, path):

        regions = []
        with open(path, 'r') as f:
            for line in f:
                line = line.strip().split('\t')
                assert(len(line) >= 3)
                regions.append(
                    cls(*line[:3], annotation= None if len(line[3:]) == 0 else line[3:])
                )
        
        return regions

    def __init__(self, chromosome, start, end, annotation = None):
        self.chromosome = str(chromosome)
        self.start = int(start)
        self.end = int(end)
        self.center = int((self.end - self.start)/ 2) + self.start
        self.annotation = annotation
        self.length = self.end - self.start

        assert(self.end > self.start), 'End position must be greater than start position'

    def to_tuple(self):
        return (self.chromosome, self.start, self.end)

    @staticmethod
    def _overlap_distance(min1, max1, min2, max2):
        return max(0, min(max1, max2) - max(min1, min2))

    def __len__(self):
        return self.end - self.start

    def overlaps(self, region, min_overlap_proportion = 0):

        assert(isinstance(region, Region))

        if self.chromosome == region.chromosome:
            overlap_dist = self._overlap_distance(self.start, self.end, region.start, region.end)
            return overlap_dist > 0 and overlap_dist >= len(self) * min_overlap_proportion
        else:
            return False

    def get_center(self):
        return self.center

    def get_genomic_distance(self, region):
        if self.chromosome == region.chromosome:
            return np.abs(self.get_center() - region.get_center())
        else:
            return np.inf

    def slop(self, d, genome):
        try:
            return Region(self.chromosome, max(0, self.start - d), min(self.end + d, genome.get_chromlen(self.chromosome)))
        except AssertionError:
            print(str(self), d)

    def __str__(self):
        return '{}\t{}\t{}'.format(self.chromosome, self.start, self.end)

    def __eq__(self, region):
        return self.chromosome == region.chromosome and self.start == region.start and self.end == region.end

    def __len__(self):
        return self.length


class Genome:

    @classmethod
    def from_file(cls, path, window_size = 100, _sort = True):
        with open(path, 'r') as f:
            chromosomes, lengths = list(zip(*[l.strip().split('\t') for l in f.readlines()]))

        return cls(chromosomes, lengths, window_size= window_size, _sort= _sort)

    def __init__(self, chromosomes, lengths, window_size = 100, _sort = True):
        
        self.window_size = int(window_size)
        self.one_bp_overlap_proportion =  1/window_size

        if _sort:
            self.chromosomes, self.lengths = list(zip(*sorted(zip(chromosomes, lengths), key = lambda x : x[0])))
        else:
            self.chromosomes, self.lengths = chromosomes, lengths
        self._sort = _sort

        self.lengths = list(map(int, self.lengths))

        self.chrom_idx = dict(zip(self.chromosomes, range(len(self.chromosomes))))

        self.indptr = [0]
        for l in self.lengths:
            self.indptr.append(self.indptr[-1] + self.get_num_windows(l, self.window_size))

    def get_chromlen(self, chromosome):

        assert(isinstance(chromosome, str))
        try:
            return self.lengths[self.chrom_idx[chromosome]]
        except KeyError:
            raise AssertionError(chromosome + ' not in genome.')

    @staticmethod
    def get_num_windows(chrom_len, window_size):
        return int(chrom_len / window_size) + int(chrom_len % window_size > 0)

    def check_region(self, region):
        if not region.chromosome in self.chrom_idx:
            raise NotInGenomeError('Chr {} from region {} not in genome'.format(str(region.chromosome), str(region)))
        else:
            try:
                assert(region.start >= 0),'Region may not start at negative index: {}'.format(str(region))
                assert(region.end <= self.get_chromlen(region.chromosome)), 'Region {} extends past end of chromosome'.format(str(region), region.chromosome)
            except AssertionError as err:
                raise BadRegionError(str(err))

    def get_window_from_position(self, chromosome, position):

        assert(isinstance(chromosome, str))
        self.check_region(Region(chromosome, position, position + 1))

        chr_start_idx = int(position / self.window_size)
        window_idx = self.indptr[self.chrom_idx[chromosome]] + chr_start_idx

        return Region(chromosome, chr_start_idx * self.window_size, min( (chr_start_idx + 1) * self.window_size, self.get_chromlen(chromosome) )), window_idx

    def get_next_window(self, region):
        assert(isinstance(region, Region))
        return self.get_window_from_position(region.chromosome, region.end)

    def num_windows_in_genome(self):
        return self.indptr[-1]

    def get_region(self, window_idx):
        assert(isinstance(window_idx, int) and window_idx >= 0)
        assert(window_idx < self.indptr[-1]), 'Invalid window idx, genome does not contain that many windows'
        for i, num_windows in enumerate(self.indptr[1:]):
            if num_windows > window_idx:
                window_idx -= self.indptr[i]
                start_pos = window_idx * self.window_size
                return self.get_window_from_position(self.chromosomes[i], start_pos)

        raise Exception('An error occured while finding window')

    def list_windows(self):

        for chromosome in self.chromosomes:
            try:
                curr_window, _ = self.get_window_from_position(chromosome, 0)
                while True:
                    yield curr_window
                    curr_window, _ = self.get_next_window(curr_window)
            except BadRegionError:
                pass

    def get_region_windows(self, region, min_window_overlap_proportion = 0.0, min_region_overlap_proportion = 0.0):

        assert(isinstance(region, Region))
        self.check_region(region)

        min_window_overlap_proportion = max(self.one_bp_overlap_proportion, min_window_overlap_proportion, 
            min(len(region)*min_region_overlap_proportion/self.window_size, 1.0))

        assert(0.0 <= min_window_overlap_proportion <= 1.0)

        min_start_loc = max(region.start - (1 - min_window_overlap_proportion) * self.window_size, 0)
        max_start_loc = max(region.end - min_window_overlap_proportion * self.window_size, 0)

        window, j = self.get_window_from_position(region.chromosome, min_start_loc)

        windows = []
        for j_plus, window_start in enumerate(range(window.start, int(max_start_loc), self.window_size)):
            if window_start >= min_start_loc:
                windows.append(j + j_plus)

        '''while window.start <= max_start_loc:

            #if window.overlaps(region, min_overlap_proportion =  min_window_overlap_proportion):
            if window.start >= min_start_loc:
                windows.append(j)
            
            try:
                window, j = self.get_next_window(window)
            except BadRegionError:
                break'''
        
        return windows
        

    def map_genomes(self, genome):

        assert(isinstance(genome, Genome))

        m2m_map = []

        for i, region in enumerate(self.list_windows()):
            try:
                windows = genome.get_region_windows(region, min_window_overlap_proportion = 0.0)
                m2m_map.extend(
                    zip([i]*len(windows), windows)
                )
            except BadRegionError:
                pass
            
        return m2m_map

    def __len__(self):
        return self.num_windows_in_genome()


class Endpoint:

    def __init__(self,point_val, endpoint_side, *, segment_obj, list_idx, list_id):
        assert(list_id in ['i','j'])
        assert(endpoint_side in ['L','R'])
        self.point_val = point_val
        self.list_id = list_id
        self.endpoint_side = endpoint_side
        self.segment_obj = segment_obj
        self.list_idx = list_idx

def get_endpoints(regions, slop_distance, genome, list_id):
    endpoints = []
    for i, region in enumerate(regions):
        slop = region.slop(slop_distance, genome)
        endpoint_args = dict(
            segment_obj = region,
            list_idx = i,
            list_id = list_id,
        )
        endpoints.append(Endpoint(slop.start, 'L', **endpoint_args))
        endpoints.append(Endpoint(slop.end,'R', **endpoint_args))
    return endpoints

def get_pairs(i_endpoints, j_endpoints, distance_function):
    interactions = []
    all_endpoints = sorted([*i_endpoints, *j_endpoints], key = lambda x : (x.point_val, x.endpoint_side))
    active_i_segments = {}
    active_j_segments = {}
    for endpoint in all_endpoints:
        if endpoint.endpoint_side == 'L':
            if endpoint.list_id == 'i':
                interactions.extend([
                    (endpoint.list_idx, segment.list_idx, distance_function(endpoint.segment_obj, segment.segment_obj))
                    for segment in active_j_segments.values()
                ])
                active_i_segments[endpoint.list_idx] = endpoint
            else:
                interactions.extend([
                    (segment.list_idx, endpoint.list_idx, distance_function(segment.segment_obj, endpoint.segment_obj))
                    for segment in active_i_segments.values()
                ])
                active_j_segments[endpoint.list_idx] = endpoint
        else:
            if endpoint.list_id == 'i':
                del active_i_segments[endpoint.list_idx]
            else:
                del active_j_segments[endpoint.list_idx]
    
    return interactions


class RegionSet:
    #stores set of genomic regions stored for quick chrom access
    def __init__(self, regions, genome):

        for region in regions:
            genome.check_region(region)

        self.genome = genome
        self.regions = sorted(regions, key = lambda r : (r.chromosome, r.start))

        self.chrom_counts = Counter([r.chromosome for r in self.regions])
        self.chrom_order = sorted(self.chrom_counts.keys())

        indptr = [0]
        self.chrom_indptr = {}
        for chrom in self.chrom_order:
            indptr.append(indptr[-1] + self.chrom_counts[chrom])
            self.chrom_indptr[chrom] = (indptr[-2], indptr[-1])

    def get_regions_by_chrom(self, chrom):
        start_idx, end_idx = self.chrom_indptr[chrom]
        return self.regions[start_idx : end_idx]

    def map_intersects(self, regions, distance_function = lambda x, y : 1, slop_distance = 0):

        assert(isinstance(slop_distance, (float, int)))
        assert(isinstance(regions, RegionSet))
        
        matrix_rows = []
        for chrom1 in self.chrom_order:
            sparse_subsections = []
            for chrom2 in regions.chrom_order:
                submatrix_shape = (self.chrom_counts[chrom1], regions.chrom_counts[chrom2])
                if chrom1 == chrom2:
                    i_endpoints = get_endpoints(self.get_regions_by_chrom(chrom1), slop_distance, self.genome, 'i')
                    j_endpoints = get_endpoints(regions.get_regions_by_chrom(chrom1), slop_distance, self.genome, 'j')
                    interaction_pairs = get_pairs(i_endpoints, j_endpoints, distance_function)
                    if len(interaction_pairs) > 0:
                        i,j,data = list(zip(*interaction_pairs))
                        sparse_subsections.append(sparse.csr_matrix((data, (i,j)), shape = submatrix_shape))
                    else:
                        sparse_subsections.append(sparse.csr_matrix(submatrix_shape))
                else:
                    sparse_subsections.append(sparse.csr_matrix(submatrix_shape))

            matrix_rows.append(sparse.hstack(sparse_subsections))

        return sparse.vstack(matrix_rows)

    def __str__(self):
        return '\n'.join([str(r) for r in self.regions])

    def __len__(self):
        return len(self.regions)

    def map_genomic_windows(self, min_window_overlap_proportion = 0.4, regions_to_bins = True):

        m2m_map = []

        for i, region in enumerate(self.regions):

            windows = self.genome.get_region_windows(region, min_window_overlap_proportion=min_window_overlap_proportion)

            if regions_to_bins:
                m2m_map.extend(
                    zip([i]*len(windows), windows)
                )
            else:
                m2m_map.extend(
                    zip(windows, [i]*len(windows))
                )

        return m2m_map




from lisa.core.utils import LoadingBar
from lisa.core.assays import LISA_RP_Assay, delta_RP_wrapper, transform_RP
from multiprocessing import Pool
import numpy as np
from scipy import stats, sparse
import os
from collections import defaultdict

class PeakRP_Assay(LISA_RP_Assay):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def predict(self, gene_mask, label_vector, data_object = None, debug = False):

        subset_factor_binding, _, subset_rp_map = self.make_subset_rp_map(
            rp_map = self.rp_map, factor_binding = self.factor_binding, 
            gene_mask = gene_mask, accessibility = self.factor_binding
        )

        rp_matrix = self.make_rp_matrix(profiles = subset_factor_binding, rp_map = subset_rp_map)
            
        self.log.append('Calculating {} peak-RP p-values ...'.format(self.technology))

        rp_matrix = np.array(rp_matrix.todense())
        
        #calculate p-values by directly applying mannu-test on RP matrix. Subset the RP matrix for genes-of-interest if required
        p_vals = self.get_delta_RP_p_value(rp_matrix, label_vector)

        query_reg_score_matrix, background_reg_score_matrix, top_factor_idx = self.get_delta_reg_score_matrix(p_vals, rp_matrix, label_vector)

        return p_vals, dict(
            reg_scores = dict(
                dataset_ids = list(np.array(self.factor_dataset_ids)[top_factor_idx]),
                query_reg_scores = query_reg_score_matrix.tolist(),
                background_reg_scores = background_reg_score_matrix.tolist()
            ),
        )

#Generator that repeatedly yields the same factor_binding and rp_map matrices with a new accessibility profile 
#to the multiprocessing pool creator. This reduces data redundancy.
class KnockoutGenerator:

    def __init__(self, accessibility_profiles, factor_binding, rp_map):
        self.accessibility_profiles = accessibility_profiles
        self.rp_map = rp_map
        self.factor_binding = factor_binding

    def __iter__(self):
        for profile in self.accessibility_profiles.T:
            yield profile, self.factor_binding, self.rp_map
    
class Accesibility_Assay(LISA_RP_Assay):

    def __init__(self, *, selection_model, chromatin_model, factor_gene_mask, cores, rp_map_style, **kwargs):
        super().__init__(**kwargs)
        self.selection_model = selection_model
        self.chromatin_model = chromatin_model
        self.factor_gene_mask = factor_gene_mask
        self.cores = cores
        self.rp_map_style = rp_map_style

    def load_accessibility_profiles(self, selected_dataset_ids):

        loadingbar = LoadingBar('Reading {} data'.format(self.technology), len(selected_dataset_ids), 20)

        accessibility_profiles = []
        metadata = dict()
        for selected_dataset in selected_dataset_ids:
            self.log.append(loadingbar, update_line = True)
            profile, sample_metadata = self.data_interface.get_profile(self.technology, selected_dataset)
            accessibility_profiles.append(profile)
            metadata.update(sample_metadata)

        accessibility_profiles = np.hstack(accessibility_profiles)

        return accessibility_profiles, metadata
    

    def calculate_ISDs(self, accessibility_profiles, factor_binding, rp_map): #enforced kwargs
        """
        accessibility_profiles: (bins, num_datasets): list of chromatin-accessiblity datasets on which to analyze the effects of ISD
        factor_binding: (num_bins, TFs), a sparse binary map showing bins that contain a chip-seq or motif peak to knock out
        rp_map: (num_bins, genes), precalculated matrix mapping the reads in a bin to a regulatory score for each gene (this is a huge matrix)

        returns:
        delta_regulatory_score (genes x TFs)
        """
        if self.cores > 1:
            self.log.append('Performing in-silico knockouts ...')
            with Pool(self.cores) as p:
                knockouts = list(p.imap(delta_RP_wrapper, iter(KnockoutGenerator(accessibility_profiles, factor_binding, rp_map))))
        else:

            bar = LoadingBar('Performing in-silico knockouts', accessibility_profiles.shape[1],
                length=20, cold_start=True)
            self.log.append(bar, update_line = True)
            knockouts = []
            for x in KnockoutGenerator(accessibility_profiles, factor_binding, rp_map):
                knockouts.append(delta_RP_wrapper(x))
                self.log.append(bar, update_line = True)


        #concatenate datasets to for gene x TF x datasets shaped matrix
        self.log.append('Calculating Δ regulatory score ...')

        datacube = np.concatenate(knockouts, axis = 1)
        num_genes, _, num_TFs = datacube.shape
        delta_regulation_score = self.chromatin_model.get_deltaRP_activation(datacube)
        
        assert(delta_regulation_score.shape == (num_genes, num_TFs))

        return delta_regulation_score

    
    def load_rp_matrix(self):
        return self.data_interface.get_rp_matrix(self.technology, self.rp_map_style)

    def introsect_accessibility(self, rp_matrix, gene_mask, dataset_mask, dataset_coefs):

        factor_accessiblities = transform_RP(rp_matrix[gene_mask, :])

        background_datasets = factor_accessiblities[:, ~dataset_mask]

        factor_means, factor_stds = background_datasets.mean(axis = 1, keepdims = True), background_datasets.std(axis = 1, keepdims = True)
        factor_acc_z_scores = (factor_accessiblities[:, dataset_mask] - factor_means)/factor_stds

        dataset_coefs = dataset_coefs.reshape(-1)
        dataset_coefs_mask = dataset_coefs > 0
        if dataset_coefs_mask.sum() == 0:
            dataset_coefs_mask = np.ones_like(dataset_coefs).astype(np.bool)

        coef_weights = dataset_coefs[dataset_coefs_mask]/dataset_coefs[dataset_coefs_mask].sum()
        factor_acc_z_scores = np.mean(np.multiply(factor_acc_z_scores[:, dataset_coefs_mask], coef_weights.reshape(1,-1)), axis = 1)

        return factor_acc_z_scores

    def predict(self, gene_mask, label_vector, debug = False):

        with self.log.section('Modeling {} purturbations:'.format(self.technology)):

            try:
                self.rp_matrix
            except AttributeError:
                self.rp_matrix, self.dataset_ids = self.load_rp_matrix()

            subset_rp_matrix = self.rp_matrix[gene_mask, :]

            #DNase model building and purturbation
            self.log.append('Selecting discriminative datasets and training chromatin model ...')

            #select the most discriminative datasets
            dataset_mask = self.selection_model.fit(subset_rp_matrix, label_vector)
            #subset the best datasets

            subset_rp_matrix, self.selected_dataset_ids = subset_rp_matrix[:, dataset_mask], self.dataset_ids[dataset_mask]
            #fit the chromatin model to these datasets
            self.chromatin_model.fit(subset_rp_matrix, label_vector, self.cores)
                    
            with self.log.section('Calculating in-silico deletions:'):

                accesibility_profiles, metadata = self.load_accessibility_profiles(self.selected_dataset_ids)

                subset_accessibility, subset_factor_binding, subset_rp_map = self.make_subset_rp_map(rp_map = self.rp_map, 
                    factor_binding = self.factor_binding, gene_mask = gene_mask, accessibility = accesibility_profiles)
                
                delta_reg_scores = self.calculate_ISDs(subset_accessibility, subset_factor_binding, subset_rp_map)

                self.log.append('Calculating p-values ...')
                
                p_vals = self.get_delta_RP_p_value(delta_reg_scores, label_vector)
        
            #self.log.append('Introspecting accessibility around factors ...')
            #factor_acc_z_scores = self.introsect_accessibility(self.rp_matrix, self.factor_gene_mask, dataset_mask,
            #    self.chromatin_model.model.coef_)
            query_reg_score_matrix, background_reg_score_matrix, top_factor_idx = self.get_delta_reg_score_matrix(p_vals, delta_reg_scores, label_vector)

            self.log.append('Done!')

        return p_vals, dict(chromatin_model = self.chromatin_model.get_info(), selection_model = self.selection_model.get_info(),
            selected_dataset_meta = self.data_interface.transpose_metadata(metadata, self.technology), #factor_acc_z_scores = factor_acc_z_scores,
            reg_scores = dict(
                dataset_ids = list(np.array(self.factor_dataset_ids)[top_factor_idx]),
                query_reg_scores = query_reg_score_matrix.tolist(),
                background_reg_scores = background_reg_score_matrix.tolist()
            ),
        )


"""
This file defines the ML models used in Lisa. The class hierarchy is shown below (* = applied directly by Lisa):

BaseEstimator --> EstimatorInterface --> SampleSelectionModel --> LR_BinarySearch_SampleSelectionModel*
                            |
                            |                   
                            --> ChromatinModel --> LR_ChromatinModel*

The EstimatorInterface enforces "fit" and "get_info" functions, and the SampleSelectionModel and ChromatinModel enforce functions
for the selection of discriminative datasets, and the training of an accurate chromatin model, respectively.

Users may extend the SampleSelectionModel or ChromatinModel class to define new behavior for Lisa.
"""

from sklearn.base import BaseEstimator
from sklearn.model_selection import GridSearchCV
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler 
from sklearn.feature_selection import SelectKBest, f_classif
import numpy as np
from sklearn.exceptions import ConvergenceWarning
import warnings
warnings.filterwarnings("ignore", category=ConvergenceWarning)
from lisa.core.assays import get_deltaRP_activation, transform_RP

class EstimatorInterface(BaseEstimator):
    #Wraps the sklearn Estimator class, and enforces that the user implement a "fit" and "get_info" method
    def fit(self):
        raise NotImplementedError("Must implement \"fit\" method of ChromatinModel")
    def get_info(self):
        raise NotImplementedError("Must implement \"get_info\" method of ChromatinModel")


class SampleSelectionModel(EstimatorInterface):
    """
    Extends the EstimatorInterface class, and enforces "get_selected_datasets" and "get_num_selected_datasets" methods
    get_selected_datasets should return a boolean index of the "selected status of each dataset in the data matrix
    """
    def get_selected_datasets(self):
        raise NotImplementedError()
    def get_num_selected_datasets(self):
        raise NotImplementedError()

class LR_BinarySearch_SampleSelectionModel(SampleSelectionModel):
    """
    Uses logistic regression to select n best discriminatory datasets. Uses binary search to tune the regularization parameter.
    max_iters: number of binary_search iterations to run, max
    num_datasets_selected: find best n datasets
    epsilon: tolerance. If dataset's regression coef is > epsilon, it is considered "selected"
    penalty: l1, use l1 penalty to choose datasets
    tol: tolerance, stop condition for LR model
    penalty_range: high and low boundaries for the C regularization parameter of the LR model.
        C = 2**-penalty, and C is inverse of regularization. Therefor for range [-1, 10], the minimum C = 2**1 = 2, very slightly regularized,
        and C = 2**-10, or very highly regularized. Binary search finds the optimal parameter in this space
    """
    def __init__(self, num_anova_features, num_datasets_selected, max_iters = 50, 
        epsilon = 1e-7, penalty = 'l1', tol = 0.01, penalty_range = (-1, 10)):
        self.max_iters = max_iters
        self.num_datasets_selected = num_datasets_selected
        self.epsilon = epsilon
        self.penalty = penalty
        self.tol = tol
        self.penalty_range = penalty_range
        self.num_anova_features = num_anova_features

    #binary search to close in on optimal penalty value
    def _binary_search(self, X, y, low, high, iter_num = 0):
        penalty = (high - low) / 2 + low
        self.model.C = 2**-penalty
        self.model.fit(X,y)
        #get upweighted datasets
        num_datasets_selected = self.get_num_selected_datasets()

        #break if the desired amount of datasets were used or max iters reached
        if num_datasets_selected == self.num_datasets_selected or iter_num == self.max_iters: 
            return self

        return self._binary_search(X, y, penalty, high, iter_num + 1) if num_datasets_selected > self.num_datasets_selected \
            else self._binary_search(X, y, low, penalty, iter_num + 1)


    #returns binary index of selected datasets
    def get_selected_datasets(self):
        return np.squeeze(np.abs(self.model.coef_) > self.epsilon)

    #returns the number of selected datasets
    def get_num_selected_datasets(self):
        return self.get_selected_datasets().sum()

    #instantiates a new LR model, then tunes the C parameter to get n datasets
    def fit(self, rp_matrix, labels):

        # Normalize with mean centering and log transformation
        index_array = np.arange(rp_matrix.shape[1])

        X = StandardScaler(with_std = False).fit_transform( transform_RP(rp_matrix) )

        #narrow features with anova selection
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            anova_featues = SelectKBest(f_classif, k=self.num_anova_features).fit(X, labels).get_support()
        #get indices mapping anova features to original features
        index_array = index_array[anova_featues]
        #subset selected features
        X = X[:, anova_featues]

        #instantiate LR search model
        self.model = LogisticRegression(penalty = self.penalty, tol = self.tol, solver = 'liblinear', random_state = 777)
        #use binary search technique to find datasets
        self._binary_search(X, labels, *self.penalty_range)
        #get boolean mask for selected datasets
        lr_selections = self.get_selected_datasets()
        #map LR selections back to original features as boolean mask
        dataset_selections = index_array[lr_selections]

        return dataset_selections

    #gets information on parameters chosen by user, and parameters of LR model to save to json logs
    def get_info(self):
        return dict(
            search_params = self.get_params(),
            search_model_params = self.model.get_params(),
            dataset_coefs = list(np.squeeze(self.model.coef_)),
        )


class ChromatinModel(EstimatorInterface):
    """
    Enforces that a chromtin model class implements get_deltaRP_activation
    """
    def get_deltaRP_activation(self, X):
        raise NotImplementedError()


class LR_ChromatinModel(ChromatinModel):

    def __init__(self, param_grid, kfold_cv = 4, scoring = 'roc_auc', penalty = 'l1'):
        self.kfold_cv = kfold_cv
        self.param_grid = param_grid
        self.scoring = scoring
        self.penalty = penalty

    def fit(self, rp_matrix, labels, n_jobs = 1):

        self.rp_0 = rp_matrix

        self.normalizer = StandardScaler(with_std = False)

        X0 = self.normalizer.fit_transform( transform_RP(self.rp_0) )

        #define a classifier for query vs background genes
        classifier = LogisticRegression(penalty = self.penalty, solver = 'lbfgs' if self.penalty == 'l2' else 'liblinear', random_state = 777)
        #optimize model with grid search
        self.grid_search = GridSearchCV(classifier, param_grid = self.param_grid, scoring = self.scoring, cv = self.kfold_cv, n_jobs = 1)\
            .fit(X0, labels)

        self.model = self.grid_search.best_estimator_

        return self

    def get_deltaRP_activation(self, rp_knockout):
        """
        rp_knockout: is a datacube of shape (genes, samples, TFs),
        this method must implement a transformation into a genes x TFs matrix, sumarrizing the dataset-axis effects
        """
        #subtract define deltaX to be the log2 of the fraction of knocked-out RP
        #deltaX = np.log2(self.rp_0[:,:,np.newaxis] - rp_knockout + PSEUDOCOUNT) - np.log2(self.rp_0[:,:,np.newaxis] + PSEUDOCOUNT)
        
        deltaX = get_deltaRP_activation(self.rp_0[:,:,np.newaxis], rp_knockout)
        
        #flip sign so that more knockout = more deltaR
        return deltaX.transpose(0,2,1).dot(self.model.coef_.reshape(-1))


    def get_info(self):
        return dict(
            model_params = self.model.get_params(),
            coefs = list(np.squeeze(self.model.coef_)),
            auc_score = self.grid_search.best_score_
        )



from urllib import request
import gzip
from lisa.core.data_interface import DataInterface
from lisa.core.genome_tools import Region, BadRegionError
from lisa.core.utils import Log, LoadingBar
import argparse
from scipy import sparse
import os
from scipy.stats import gamma
import numpy as np
from sys import stdout, stderr

def main(species, motif_bed, window_size, gamma_threshold = 0.95):

    genome = DataInterface.load_genome(species, window_size)

    log = Log(target= stderr)
    
    factor_name = None
    window_nums, scores = [],[]

    with gzip.open(motif_bed, 'rb') as f:

        bed = f.readlines()

        bar = LoadingBar('Binning {} motif hits'.format(str(len(bed))), len(bed), cold_start=True)

        for i, line in enumerate(bed):
            
            chrom, start, end, factor, relscore, log_pval, strand = line.decode('utf-8').strip().split('\t')
            
            if i == 0:
                factor_name = factor

            try:
                hit_windows = genome.get_region_windows(Region(chrom, start, end))
                window_nums.extend(hit_windows)

                scores.extend([float(log_pval)/100]*len(hit_windows))

            except BadRegionError:
                pass

            log.append(bar, update_line=True)
    
    log.append('')
        
    log.append('Done')

    hits = sparse.csc_matrix((scores, window_nums, [0, len(window_nums)]), shape = (len(genome), 1)).tocoo().tocsc()

    sample_hit_scores = np.random.choice(np.array(hits.todense()).reshape(-1), size = 10000)

    min_bin_score = gamma(*gamma.fit(sample_hit_scores)).ppf(gamma_threshold)

    hit_indices = hits.indices[(hits.data >= min_bin_score) & (hits.data > 0)]
    
    return hit_indices, factor_name


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument('species', type = str, choices = ['hg38','mm10'])
    parser.add_argument('motif_bed', type = str)
    parser.add_argument('-w', '--window_size', type = int, required=True)
    parser.add_argument('-o', '--output', type = str, required=True)
    args = parser.parse_args()

    hit_indices, factor_name = main(args.species, args.motif_bed, args.window_size)

    motif_id = '.'.join(os.path.basename(args.motif_bed).split('.')[:2])

    print(motif_id, factor_name, 'JASPAR', sep = '\t', file = stdout)

    with open(args.output, 'w') as f:
        print('\n'.join(map(str, hit_indices)), file = f)    




from setuptools import setup, find_packages

setup(
    name = 'lisa2',
    description = "Lisa: inferring transcriptional regulators through integrative modeling of public chromatin accessibility and ChIP-seq data. X. Shirley Liu Lab, 2020",
    version = '2.2.3',
    url = 'https://genomebiology.biomedcentral.com/articles/10.1186/s13059-020-1934-6',
    author = 'Allen Lynch',
    author_email = 'alynch@ds.dfci.harvard.edu',
    packages = find_packages(),
    zip_safe = True,
    scripts = ['bin/lisa'],
    install_requires = [
        'numpy>=1.19,<2',
        'scipy>=1.5,<2',
        'h5py>=2.10.0,<3',
        'scikit-learn>=0.23.2,<1',
        'pyBigWig>=0.3.17,<1'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    python_requires='>=3.6,<4',
    long_description_content_type='text/markdown',
    long_description = '''
# About

LISA is a statistical test for the influence of Transcription Factors on a set of genes. We leverage integrative modeling of public chromatin accessiblity and factor binding to 
make predictions that go beyond simple co-expression analysis. The minimum you need to run LISA is a list of genes-of-interest, 
but you can also supply your own epigenetic background. For more information, see Qin et al., 2020. 
This implementation extends the original, running faster, reducing dependencies, and adding useful CLI functions for pipeline integration.

# Documentation

Please see Lisa's [github repo](https://github.com/liulab-dfci/lisa2) for source code and tutorials.

# Installation

### Requirements

* Mac or Linux OS
* Python 3.6+
* 15 GB of available storage space

### Installation

LISA will install data into the virutal environment's "site_packages" directory, so ensure the env's location can store ~10GB.

### PyPI

It is recommended to install lisa to a virtual environment:

    $ python3 -m venv .venvs/lisa_env
    $ source .venvs/lisa_env/bin/activate

Install LISA to this virtual env using this command:

    (lisa_env) $ pip install lisa2

### Conda

First, create a virtual environment:

    (base) $ conda create --name lisa_env
    (base) $ conda activate lisa_env

Then install from Conda:

    (lisa_env) $ conda install -c liulab-dfci lisa2

### Dependencies

* numpy
* scipy
* h5py
* pyBigWig
* scikit-learn

    ''',
)



from lisa.core.io import parse_bedfile
from lisa.core.data_interface import DataInterface
from lisa.core.genome_tools import Region, BadRegionError
import argparse

def main(species, window_size, path):

    region_fields = parse_bedfile(path, header = False)

    regions = [Region(*r) for r in region_fields]

    genome = DataInterface.load_genome(species, window_size)

    indices = []

    for region in regions:
        try:
            windows = genome.get_region_windows(region)
            indices.extend(windows)
        except BadRegionError:
            pass

    return list(set(indices))

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Using summit file or motif hits file, find genome indices where regions overlap')
    parser.add_argument('species', type = str, choices = ['hg38','mm10'])
    parser.add_argument('window_size', type = int)
    parser.add_argument('input', type = str)
    parser.add_argument('output', type = str)

    args = parser.parse_args()

    indices = main(args.species, int(args.window_size), args.input)

    with open(args.output, 'w') as f:
        print('\n'.join(map(str, indices)), file = f)